@import url('https://rsms.me/inter/inter.css');
@import url('https://assets.hynek.me/css/bm.css');


:root {
  font-feature-settings: 'liga' 1, 'calt' 1; /* fix for Chrome */
}
@supports (font-variation-settings: normal) {
  :root { font-family: InterVariable, sans-serif; }
}
